/*
 * Copyright 2013-2014 aquenos GmbH.
 * All rights reserved.
 * 
 * This program and the accompanying materials are made available under the 
 * terms of the Eclipse Public License v1.0 which accompanies this distribution,
 * and is available at http://www.eclipse.org/legal/epl-v10.html.
 */

package com.aquenos.csstudio.archive.json.server;

import java.io.IOException;
import java.io.OutputStream;
import java.math.BigDecimal;
import java.math.BigInteger;
import java.net.URLDecoder;
import java.text.NumberFormat;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import java.util.zip.DeflaterOutputStream;
import java.util.zip.GZIPOutputStream;

import javax.servlet.ServletException;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.codehaus.jackson.JsonFactory;
import org.codehaus.jackson.JsonGenerator;
import org.csstudio.archive.reader.ArchiveInfo;
import org.csstudio.archive.reader.ArchiveReader;
import org.csstudio.archive.reader.ArchiveReaderFactory;
import org.csstudio.archive.reader.ArchiveRepository;
import org.csstudio.archive.reader.UnknownChannelException;
import org.csstudio.archive.reader.ValueIterator;
import org.epics.util.array.ArrayByte;
import org.epics.util.array.ArrayDouble;
import org.epics.util.array.ArrayFloat;
import org.epics.util.array.ArrayInt;
import org.epics.util.array.ArrayLong;
import org.epics.util.array.ArrayShort;
import org.epics.util.array.IteratorNumber;
import org.epics.util.array.ListByte;
import org.epics.util.array.ListDouble;
import org.epics.util.array.ListFloat;
import org.epics.util.array.ListInt;
import org.epics.util.array.ListLong;
import org.epics.util.array.ListNumber;
import org.epics.util.array.ListShort;
import org.epics.util.time.Timestamp;
import org.epics.vtype.Alarm;
import org.epics.vtype.AlarmSeverity;
import org.epics.vtype.Display;
import org.epics.vtype.Statistics;
import org.epics.vtype.Time;
import org.epics.vtype.VEnum;
import org.epics.vtype.VEnumArray;
import org.epics.vtype.VNumber;
import org.epics.vtype.VNumberArray;
import org.epics.vtype.VString;
import org.epics.vtype.VStringArray;
import org.epics.vtype.VType;

/**
 * Servlet that exposes an {@link ArchiveReader} through a JSON-based protocol.
 * 
 * @author Sebastian Marsching
 */
public class ArchiveServlet extends HttpServlet {

    private static final long serialVersionUID = -1272334811448099566L;

    private static final BigInteger ONE_BILLION = BigInteger
            .valueOf(1000000000L);

    private ArchiveReader archiveReader;

    private final JsonFactory jsonFactory;

    /**
     * Constructor. Creates an instance of the archive servlet, that server
     * requests for a specific archive URL.
     * 
     * @param archiveUrl
     *            the archive URL. The URL must be supported by the
     *            {@link ArchiveReaderFactory}.
     * @throws Exception
     *             if archive URL is not supported or the
     *             {@link ArchiveReaderFactory} throws an exception.
     */
    public ArchiveServlet(String archiveUrl) throws Exception {
        super();
        archiveReader = ArchiveRepository.getInstance().getArchiveReader(
                archiveUrl);
        jsonFactory = new JsonFactory();
    }

    @Override
    protected void doGet(HttpServletRequest request,
            HttpServletResponse response) throws ServletException, IOException {
        String path = request.getPathInfo();
        if (path == null || path.equals("/")) {
            generateArchiveInfo(request, response);
        } else {
            String[] pathComponents = path.substring(1).split("/", 3);
            if (pathComponents.length < 3) {
                response.sendError(HttpServletResponse.SC_NOT_FOUND);
            } else {
                String archiveIdString = pathComponents[0];
                int archiveId;
                try {
                    archiveId = Integer.parseInt(archiveIdString);
                } catch (NumberFormatException e) {
                    response.sendError(HttpServletResponse.SC_BAD_REQUEST);
                    return;
                }
                String action = pathComponents[1];
                String actionArgument = URLDecoder.decode(pathComponents[2],
                        "UTF-8");
                if (action.equals("channels-by-pattern")) {
                    try {
                        generateChannelsList(request, response,
                                archiveReader.getNamesByPattern(archiveId,
                                        actionArgument));
                    } catch (Exception e) {
                        response.sendError(HttpServletResponse.SC_INTERNAL_SERVER_ERROR);
                    }
                } else if (action.equals("channels-by-regexp")) {
                    try {
                        generateChannelsList(request, response,
                                archiveReader.getNamesByRegExp(archiveId,
                                        actionArgument));
                    } catch (Exception e) {
                        response.sendError(HttpServletResponse.SC_INTERNAL_SERVER_ERROR);
                    }
                } else if (action.equals("samples")) {
                    Timestamp start = stringToTimestamp(request
                            .getParameter("start"));
                    Timestamp end = stringToTimestamp(request
                            .getParameter("end"));
                    String countString = request.getParameter("count");
                    if (start == null || end == null) {
                        response.sendError(HttpServletResponse.SC_BAD_REQUEST);
                        return;
                    }
                    if (countString == null) {
                        try {
                            generateSamplesList(request, response,
                                    archiveReader.getRawValues(archiveId,
                                            actionArgument, start, end));
                        } catch (UnknownChannelException e) {
                            response.sendError(HttpServletResponse.SC_NOT_FOUND);
                        } catch (Exception e) {
                            response.sendError(HttpServletResponse.SC_INTERNAL_SERVER_ERROR);
                        }
                    } else {
                        int count;
                        try {
                            count = Integer.valueOf(countString);
                        } catch (NumberFormatException e) {
                            response.sendError(HttpServletResponse.SC_BAD_REQUEST);
                            return;
                        }
                        try {
                            generateSamplesList(request, response,
                                    archiveReader.getOptimizedValues(archiveId,
                                            actionArgument, start, end, count));
                        } catch (UnknownChannelException e) {
                            response.sendError(HttpServletResponse.SC_NOT_FOUND);
                        } catch (Exception e) {
                            response.sendError(HttpServletResponse.SC_INTERNAL_SERVER_ERROR);
                        }
                    }
                } else {
                    response.sendError(HttpServletResponse.SC_NOT_FOUND);
                }
            }
        }
    }

    private static boolean prettyPrintRequested(HttpServletRequest request) {
        String prettyPrintParameter = request.getParameter("prettyPrint");
        if (prettyPrintParameter != null) {
            return true;
        } else {
            return false;
        }
    }

    private static void enablePrettyPrintIfRequested(
            HttpServletRequest request, JsonGenerator generator) {
        if (prettyPrintRequested(request)) {
            generator.useDefaultPrettyPrinter();
        }
    }

    private static Timestamp stringToTimestamp(String string) {
        if (string == null) {
            return null;
        }
        BigInteger timestampInNanoSeconds;
        try {
            timestampInNanoSeconds = new BigInteger(string);
        } catch (NumberFormatException e) {
            return null;
        }
        BigInteger[] quotientAndRemainder = timestampInNanoSeconds
                .divideAndRemainder(ONE_BILLION);
        return Timestamp.of(quotientAndRemainder[0].longValue(),
                (int) quotientAndRemainder[1].longValue());
    }

    private void generateSamplesList(HttpServletRequest request,
            HttpServletResponse response, ValueIterator valueIterator)
            throws IOException {
        response.setCharacterEncoding("UTF-8");
        response.setContentType("application/json");
        JsonGenerator generator = jsonFactory
                .createJsonGenerator(getCompressedOutputStream(request,
                        response));
        enablePrettyPrintIfRequested(request, generator);
        try {
            generator.writeStartArray();
            while (valueIterator.hasNext()) {
                try {
                    VType value = valueIterator.next();
                    writeValue(generator, value);
                } catch (Exception e) {
                    // We first flush the generator and then close the stream
                    // before closing the generator. This has the effect that
                    // the JSON stream is not finished (no closed array) which
                    // is the best way to signal to the client that an error
                    // occurred while sending the samples.
                    generator.flush();
                    response.getOutputStream().close();
                    return;
                }
            }
            generator.writeEndArray();
        } finally {
            generator.close();
        }
    }

    private void writeValue(JsonGenerator generator, VType value)
            throws IOException {
        generator.writeStartObject();
        Timestamp timestamp = null;
        if (value instanceof Time) {
            Time time = (Time) value;
            timestamp = time.getTimestamp();
        }
        if (timestamp != null) {
            generator.writeNumberField("time", new BigDecimal(
                    timestampToBigInteger(timestamp)));
        } else {
            generator.writeNumberField("time", 0);
        }
        generator.writeFieldName("severity");
        String status = null;
        AlarmSeverity severity = null;
        if (value instanceof Alarm) {
            Alarm alarm = (Alarm) value;
            status = alarm.getAlarmName();
            severity = alarm.getAlarmSeverity();
        }
        if (status == null) {
            status = "";
        }
        if (severity == null) {
            severity = AlarmSeverity.NONE;
        } else if (severity == AlarmSeverity.UNDEFINED) {
            // We cannot use the undefined severity because it is not supported
            // by this protocol version.
            severity = AlarmSeverity.NONE;
        }
        writeSeverity(generator, severity);
        generator.writeStringField("status", status);
        generator.writeFieldName("quality");
        if (value instanceof Statistics) {
            generator.writeString("Interpolated");
        } else {
            generator.writeString("Original");
        }
        if (value instanceof VEnum) {
            VEnum enumValue = (VEnum) value;
            List<String> labels = enumValue.getLabels();
            if (labels == null) {
                labels = Collections.emptyList();
            }
            generator.writeFieldName("metaData");
            writeEnumeratedMetaData(generator, labels);
        } else if (value instanceof VEnumArray) {
            VEnumArray enumValue = (VEnumArray) value;
            List<String> labels = enumValue.getLabels();
            if (labels == null) {
                labels = Collections.emptyList();
            }
            generator.writeFieldName("metaData");
            writeEnumeratedMetaData(generator, labels);
        } else if (value instanceof Display) {
            Display display = (Display) value;
            NumberFormat numberFormat = display.getFormat();
            int precision = 0;
            if (numberFormat != null) {
                precision = numberFormat.getMaximumFractionDigits();
            }
            Double alarmLow = display.getLowerAlarmLimit();
            if (alarmLow == null) {
                alarmLow = 0.0;
            }
            Double alarmHigh = display.getUpperAlarmLimit();
            if (alarmHigh == null) {
                alarmHigh = 0.0;
            }
            Double warnLow = display.getLowerWarningLimit();
            if (warnLow == null) {
                warnLow = 0.0;
            }
            Double warnHigh = display.getUpperWarningLimit();
            if (warnHigh == null) {
                warnHigh = 0.0;
            }
            Double displayLow = display.getLowerDisplayLimit();
            if (displayLow == null) {
                displayLow = 0.0;
            }
            Double displayHigh = display.getUpperDisplayLimit();
            if (displayHigh == null) {
                displayHigh = 0.0;
            }
            String units = display.getUnits();
            if (units == null) {
                units = "";
            }
            generator.writeFieldName("metaData");
            writeNumericMetaData(generator, precision, units, displayLow,
                    displayHigh, alarmLow, alarmHigh, warnLow, warnHigh);
        }
        if (value instanceof Statistics) {
            generator.writeStringField("type", "minMaxDouble");
            Statistics statistics = (Statistics) value;
            generator.writeFieldName("value");
            if (value instanceof VNumberArray) {
                VNumberArray numberArray = (VNumberArray) value;
                ListNumber listNumber = numberArray.getData();
                if (listNumber == null) {
                    listNumber = new ArrayDouble();
                }
                writeNumberValue(generator, listNumber);
            } else if (value instanceof VNumber) {
                VNumber numberValue = (VNumber) value;
                Number number = numberValue.getValue();
                if (number != null) {
                    if (number instanceof Byte) {
                        writeNumberValue(generator,
                                new ArrayByte(number.byteValue()));
                    } else if (number instanceof Double) {
                        writeNumberValue(generator,
                                new ArrayDouble(number.doubleValue()));
                    } else if (number instanceof Float) {
                        writeNumberValue(generator,
                                new ArrayFloat(number.floatValue()));
                    } else if (number instanceof Integer) {
                        writeNumberValue(generator,
                                new ArrayInt(number.intValue()));
                    } else if (number instanceof Long) {
                        writeNumberValue(generator,
                                new ArrayLong(number.longValue()));
                    } else if (number instanceof Short) {
                        writeNumberValue(generator,
                                new ArrayShort(number.shortValue()));
                    } else {
                        throw new RuntimeException(
                                "Got java.lang.Number instance of unexpected type "
                                        + number.getClass().getName() + ".");
                    }
                }
            } else {
                Double doubleValue = statistics.getAverage();
                if (doubleValue == null) {
                    writeNumberValue(generator, new ArrayDouble());
                } else {
                    writeNumberValue(generator, new ArrayDouble(doubleValue));
                }
            }
            Double minimum = statistics.getMin();
            if (minimum == null) {
                minimum = Double.NEGATIVE_INFINITY;
            }
            Double maximum = statistics.getMax();
            if (maximum == null) {
                maximum = Double.POSITIVE_INFINITY;
            }
            generator.writeNumberField("minimum", minimum);
            generator.writeNumberField("maximum", maximum);
        } else if (value instanceof VNumberArray) {
            VNumberArray numberArray = (VNumberArray) value;
            ListNumber listNumber = numberArray.getData();
            if (listNumber == null) {
                listNumber = new ArrayDouble();
            }
            if (listNumber instanceof ListByte
                    || listNumber instanceof ListShort
                    || listNumber instanceof ListInt
                    || listNumber instanceof ListLong) {
                generator.writeStringField("type", "long");
            } else {
                generator.writeStringField("type", "double");
            }
            generator.writeFieldName("value");
            writeNumberValue(generator, listNumber);
        } else if (value instanceof VNumber) {
            VNumber numberValue = (VNumber) value;
            Number number = numberValue.getValue();
            if (number != null) {
                if (number instanceof Byte) {
                    generator.writeStringField("type", "long");
                    generator.writeFieldName("value");
                    writeNumberValue(generator,
                            new ArrayByte(number.byteValue()));
                } else if (number instanceof Double) {
                    generator.writeStringField("type", "double");
                    generator.writeFieldName("value");
                    writeNumberValue(generator,
                            new ArrayDouble(number.doubleValue()));
                } else if (number instanceof Float) {
                    generator.writeStringField("type", "double");
                    generator.writeFieldName("value");
                    writeNumberValue(generator,
                            new ArrayFloat(number.floatValue()));
                } else if (number instanceof Integer) {
                    generator.writeStringField("type", "long");
                    generator.writeFieldName("value");
                    writeNumberValue(generator, new ArrayInt(number.intValue()));
                } else if (number instanceof Long) {
                    generator.writeStringField("type", "long");
                    generator.writeFieldName("value");
                    writeNumberValue(generator,
                            new ArrayLong(number.longValue()));
                } else if (number instanceof Short) {
                    generator.writeStringField("type", "long");
                    generator.writeFieldName("value");
                    writeNumberValue(generator,
                            new ArrayShort(number.shortValue()));
                } else {
                    // Treat other numbers as doubles.
                    generator.writeStringField("type", "double");
                    generator.writeFieldName("value");
                    writeNumberValue(generator,
                            new ArrayDouble(number.doubleValue()));
                }
            }
        } else if (value instanceof VEnumArray) {
            VEnumArray enumArray = (VEnumArray) value;
            ListInt indexes = enumArray.getIndexes();
            if (indexes == null) {
                indexes = new ArrayInt();
            }
            generator.writeStringField("type", "enum");
            generator.writeFieldName("value");
            writeNumberValue(generator, indexes);
        } else if (value instanceof VEnum) {
            VEnum enumValue = (VEnum) value;
            generator.writeStringField("type", "enum");
            generator.writeFieldName("value");
            writeNumberValue(generator, new ArrayInt(enumValue.getIndex()));
        } else if (value instanceof VStringArray) {
            VStringArray stringArray = (VStringArray) value;
            List<String> stringList = stringArray.getData();
            if (stringList == null) {
                stringList = Collections.emptyList();
            }
            generator.writeStringField("type", "string");
            generator.writeFieldName("value");
            writeStringValue(generator, stringList);
        } else if (value instanceof VString) {
            VString stringValue = (VString) value;
            String string = stringValue.getValue();
            generator.writeStringField("type", "string");
            generator.writeFieldName("value");
            if (string == null) {
                writeStringValue(generator, Collections.<String> emptyList());
            } else {
                writeStringValue(generator, Arrays.asList(string));
            }
        } else {
            generator.writeStringField("type", "string");
            generator.writeFieldName("value");
            writeStringValue(generator, Arrays.asList("Unsupported VType."));
        }
        generator.writeEndObject();
    }

    private void writeSeverity(JsonGenerator generator, AlarmSeverity severity)
            throws IOException {
        generator.writeStartObject();
        switch (severity) {
        case NONE:
            generator.writeStringField("level", "OK");
            break;
        case MINOR:
            generator.writeStringField("level", "MINOR");
            break;
        case MAJOR:
            generator.writeStringField("level", "MAJOR");
            break;
        case INVALID:
            generator.writeStringField("level", "INVALID");
            break;
        default:
            // We use the OK severity is a default because the protocol does not
            // support the UNDEFINED severity yet (this would break old
            // clients).
            generator.writeStringField("level", "OK");
        }
        generator.writeBooleanField("hasValue", true);
        generator.writeEndObject();
    }

    private void writeEnumeratedMetaData(JsonGenerator generator,
            List<String> labels) throws IOException {
        generator.writeStartObject();
        generator.writeStringField("type", "enum");
        generator.writeFieldName("states");
        generator.writeStartArray();
        for (String label : labels) {
            generator.writeString(label);
        }
        generator.writeEndArray();
        generator.writeEndObject();
    }

    private void writeNumericMetaData(JsonGenerator generator, int precision,
            String units, double displayLow, double displayHigh,
            double alarmLow, double alarmHigh, double warnLow, double warnHigh)
            throws IOException {
        generator.writeStartObject();
        generator.writeStringField("type", "numeric");
        generator.writeNumberField("precision", precision);
        generator.writeStringField("units", units);
        generator.writeNumberField("displayLow", displayLow);
        generator.writeNumberField("displayHigh", displayHigh);
        generator.writeNumberField("warnLow", warnLow);
        generator.writeNumberField("warnHigh", warnHigh);
        generator.writeNumberField("alarmLow", alarmLow);
        generator.writeNumberField("alarmHigh", alarmHigh);
        generator.writeEndObject();
    }

    private void writeNumberValue(JsonGenerator generator, ListNumber numbers)
            throws IOException {
        generator.writeStartArray();
        if (numbers instanceof ListByte) {
            for (IteratorNumber i = numbers.iterator(); i.hasNext();) {
                generator.writeNumber(i.nextByte());
            }
        } else if (numbers instanceof ListDouble) {
            for (IteratorNumber i = numbers.iterator(); i.hasNext();) {
                generator.writeNumber(i.nextDouble());
            }
        } else if (numbers instanceof ListFloat) {
            for (IteratorNumber i = numbers.iterator(); i.hasNext();) {
                generator.writeNumber(i.nextFloat());
            }
        } else if (numbers instanceof ListInt) {
            for (IteratorNumber i = numbers.iterator(); i.hasNext();) {
                generator.writeNumber(i.nextInt());
            }
        } else if (numbers instanceof ListLong) {
            for (IteratorNumber i = numbers.iterator(); i.hasNext();) {
                generator.writeNumber(i.nextLong());
            }
        } else if (numbers instanceof ListShort) {
            for (IteratorNumber i = numbers.iterator(); i.hasNext();) {
                generator.writeNumber(i.nextShort());
            }
        } else {
            // Treat everything else as doubles.
            for (IteratorNumber i = numbers.iterator(); i.hasNext();) {
                generator.writeNumber(i.nextDouble());
            }
        }
        generator.writeEndArray();
    }

    private void writeStringValue(JsonGenerator generator, List<String> value)
            throws IOException {
        generator.writeStartArray();
        for (String s : value) {
            generator.writeString(s);
        }
        generator.writeEndArray();
    }

    private BigInteger timestampToBigInteger(Timestamp timestamp) {
        return BigInteger.valueOf(timestamp.getNanoSec()).add(
                BigInteger.valueOf(timestamp.getSec()).multiply(ONE_BILLION));
    }

    private void generateChannelsList(HttpServletRequest request,
            HttpServletResponse response, String[] channelNames)
            throws IOException {
        response.setCharacterEncoding("UTF-8");
        response.setContentType("application/json");
        JsonGenerator generator = jsonFactory
                .createJsonGenerator(getCompressedOutputStream(request,
                        response));
        enablePrettyPrintIfRequested(request, generator);
        try {
            generator.writeStartArray();
            for (String channelName : channelNames) {
                generator.writeString(channelName);
            }
            generator.writeEndArray();
        } finally {
            generator.close();
        }
    }

    private void generateArchiveInfo(HttpServletRequest request,
            HttpServletResponse response) throws IOException {
        response.setCharacterEncoding("UTF-8");
        response.setContentType("application/json");
        JsonGenerator generator = jsonFactory
                .createJsonGenerator(getCompressedOutputStream(request,
                        response));
        enablePrettyPrintIfRequested(request, generator);
        try {
            generator.writeStartArray();
            for (ArchiveInfo archiveInfo : archiveReader.getArchiveInfos()) {
                generator.writeStartObject();
                generator.writeNumberField("key", archiveInfo.getKey());
                generator.writeStringField("name", archiveInfo.getName());
                generator.writeStringField("description",
                        archiveInfo.getDescription());
                generator.writeEndObject();
            }
            generator.writeEndArray();
        } finally {
            generator.close();
        }
    }

    private OutputStream getCompressedOutputStream(HttpServletRequest request,
            HttpServletResponse response) throws IOException {
        String acceptEncodingHeader = request.getHeader("Accept-Encoding");
        if (acceptEncodingHeader == null) {
            return response.getOutputStream();
        }
        String[] acceptedEncodings = acceptEncodingHeader.split(",");
        boolean supportsGzip = false;
        boolean supportsDeflate = false;
        for (String encoding : acceptedEncodings) {
            encoding = encoding.trim();
            if (encoding.trim().equals("gzip")) {
                supportsGzip = true;
            } else if (encoding.trim().equals("deflate")) {
                supportsDeflate = true;
            }
        }
        if (supportsGzip) {
            response.setHeader("Content-Encoding", "gzip");
            return new GZIPOutputStream(response.getOutputStream(), 65536);
        } else if (supportsDeflate) {
            response.setHeader("Content-Encoding", "deflate");
            return new DeflaterOutputStream(response.getOutputStream());
        } else {
            return response.getOutputStream();
        }
    }
}
